@@ -1,22 +1,21 @@
 ;; move-resize.jl -- interactive moving and resizing of windows
-;; $Id: move-resize.jl,v 1.91 2002/05/29 06:39:02 jsh Exp $
 
 ;; Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
-;; This file is part of sawmill.
+;; This file is part of sawfish.
 
-;; sawmill is free software; you can redistribute it and/or modify it
+;; sawfish is free software; you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; sawmill is distributed in the hope that it will be useful, but
+;; sawfish is distributed in the hope that it will be useful, but
 ;; WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with sawmill; see the file COPYING.  If not, write to
+;; along with sawfish; see the file COPYING.  If not, write to
 ;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
 (define-structure sawfish.wm.commands.move-resize
@@ -24,7 +23,10 @@
     (export move-window-interactively
 	    resize-window-interactively
 	    move-selected-window
-	    resize-selected-window)
+	    resize-selected-window
+	    double-window-size
+	    halve-window-size
+	    move-window-center)
 
     (open rep
 	  rep.system
@@ -48,60 +50,69 @@
   ;; todo:
   ;;  * obey the aspect ratio size hints
 
-  (defvar move-outline-mode 'opaque
-    "How windows being moved are animated, either `opaque' or `box'")
-  
-  (defvar resize-outline-mode 'opaque
-    "How windows being resized are animated, either `opaque' or `box'")
-  
-  (defvar move-resize-raise-window nil
-    "Raise windows when they are moved or resized.")
-  
+
+  (defcustom move-outline-mode 'opaque
+    "How windows being moved are animated"
+    :type (choice opaque box cross elliptical draft)
+    :group (appearance animation))
+
+  (defcustom resize-outline-mode 'opaque
+    "How windows being resized are animated"
+    :type (choice opaque box cross elliptical draft)
+    :group (appearance animation))
+
+  (defcustom move-resize-raise-window nil
+    "Raise windows when they are moved or resized."
+    :type boolean
+    :group move)
+
   (defcustom move-show-position nil
     "Show current position of windows while moving."
     :group move
     :type boolean)
-  
+
   (defcustom resize-show-dimensions t
     "Show current dimensions of windows while resizing."
     :group move
     :type boolean)
-  
-  (defvar resize-edge-mode 'border-grab
-    "How to choose window edges when resizing. One of `region', `border',
-`grab', `border-grab'")
-  
+
+  (defcustom resize-edge-mode 'border-grab
+    "How to choose window edges when resizing."
+    :type (choice region border grab border-grab)
+    :group move)
+
   (defcustom move-snap-epsilon 12
     "Distance in pixels before window edges align with each other."
     :group move
     :type (number 0 64)
-    :tooltip "When moving a window, this option lets you align one of its edges with an edge of another window.")
-  
+    :tooltip "When moving a window, this option lets you align one of \
+its edges with an edge of another window.")
+
   (defvar move-snap-mode 'resistance
     "How to snap together window edges, one of `magnetism', `resistance', or
 `attraction'.")
-  
+
   (defvar move-snap-ignored-windows nil
     "Snap to otherwise-ignored windows.")
-  
+
   (defvar move-resize-inhibit-configure nil
     "Only update window contents after it has stopped moving.")
 
   (defvar move-resize-map (bind-keys (make-keymap)
-			    "Any-Off1" (lambda () (finished))
-			    "Any-Off2" (lambda () (finished))
-			    "Any-Off3" (lambda () (finished))
-			    "Any-Move" (lambda () (motion))
-			    "Any-ESC" (lambda () (cancel))
-			    "Any-RET" (lambda () (finished))
-			    "Up" 'move-cursor-up
-			    "Down" 'move-cursor-down
-			    "Left" 'move-cursor-left
-			    "Right" 'move-cursor-right
-			    "S-Up" 'move-cursor-up-fine
-			    "S-Down" 'move-cursor-down-fine
-			    "S-Left" 'move-cursor-left-fine
-			    "S-Right" 'move-cursor-right-fine))
+                                     "Any-Off1" (lambda () (finished))
+                                     "Any-Off2" (lambda () (finished))
+                                     "Any-Off3" (lambda () (finished))
+                                     "Any-Move" (lambda () (motion))
+                                     "Any-ESC" (lambda () (cancel))
+                                     "Any-RET" (lambda () (finished))
+                                     "Up" 'move-cursor-up
+                                     "Down" 'move-cursor-down
+                                     "Left" 'move-cursor-left
+                                     "Right" 'move-cursor-right
+                                     "S-Up" 'move-cursor-up-fine
+                                     "S-Down" 'move-cursor-down-fine
+                                     "S-Left" 'move-cursor-left-fine
+                                     "S-Right" 'move-cursor-right-fine))
 
   (define fp-edges-alist '((top-border top)
 			   (left-border left)
@@ -112,8 +123,8 @@
 			   (bottom-left-corner bottom left)
 			   (bottom-right-corner bottom right)))
 
-  (defvar move-cursor-shape (default-cursor))
-  (defvar resize-cursor-shape (default-cursor))
+  (defvar move-cursor-shape 'hand2)
+  (defvar resize-cursor-shape 'crosshair)
 
   ;; specials, should make these fluids (external users?)
   (defvar move-resize-window nil)
@@ -308,7 +319,10 @@
 		 ((x-base (or (cdr (assq 'base-width move-resize-hints)) 0))
 		  (x-inc (or (cdr (assq 'width-inc move-resize-hints)) 1))
 		  (y-base (or (cdr (assq 'base-height move-resize-hints)) 0))
-		  (y-inc (or (cdr (assq 'height-inc move-resize-hints)) 1)))
+		  (y-inc (or (cdr (assq 'height-inc move-resize-hints)) 1))
+		  (min-aspect (assq 'min-aspect move-resize-hints))
+		  (max-aspect (assq 'max-aspect move-resize-hints)))
+
 	       (when (memq resize-edge-mode '(grab border-grab))
 		 (add-edges ptr-x ptr-y))
 	       (cond
@@ -317,13 +331,24 @@
 		       (constrain-dimension-to-hints
 			(+ move-resize-old-width
 			   (- ptr-x move-resize-old-ptr-x))
-			'x move-resize-hints)))
+			'x move-resize-hints))
+                 (when (or min-aspect max-aspect)
+                   (setq move-resize-height
+                         (constrain-aspect-to-hints
+                          move-resize-width
+                          move-resize-old-height 'x min-aspect max-aspect))))
+
 		((memq 'left move-resize-moving-edges)
 		 (setq move-resize-width
 		       (constrain-dimension-to-hints
 			(+ move-resize-old-width
 			   (- move-resize-old-ptr-x ptr-x))
-			'x move-resize-hints))
+                        'x move-resize-hints))
+                 (when (or min-aspect max-aspect)
+                   (setq move-resize-height
+                         (constrain-aspect-to-hints
+                          move-resize-width
+                          move-resize-old-height 'x min-aspect max-aspect)))
 		 (setq move-resize-x (- move-resize-old-x
 					(- move-resize-width
 					   move-resize-old-width)))))
@@ -333,13 +358,24 @@
 		       (constrain-dimension-to-hints
 			(+ move-resize-old-height
 			   (- ptr-y move-resize-old-ptr-y))
-			'y move-resize-hints)))
+			'y move-resize-hints))
+                 (when (or min-aspect max-aspect)
+                   (setq move-resize-width
+                         (constrain-aspect-to-hints
+                          move-resize-height
+                          move-resize-old-width 'y min-aspect max-aspect)))
+                 )
 		((memq 'top move-resize-moving-edges)
 		 (setq move-resize-height
 		       (constrain-dimension-to-hints
 			(+ move-resize-old-height
 			   (- move-resize-old-ptr-y ptr-y))
 			'y move-resize-hints))
+                 (when (or min-aspect max-aspect)
+                   (setq move-resize-width
+                         (constrain-aspect-to-hints
+                          move-resize-height
+                          move-resize-old-width 'y min-aspect max-aspect)))
 		 (setq move-resize-y (- move-resize-old-y
 					(- move-resize-height
 					   move-resize-old-height)))))
@@ -480,7 +516,22 @@
 		      (copy-sequence move-resize-moving-edges)
 		    (list 'top 'bottom 'left 'right)))))))
 
-
+  (define (resize-by-factor win amount)
+    "Multiply win's dimensions by amount"
+    (let* (
+           (orig-wid (car (window-dimensions win)))
+           (orig-hgt (cdr (window-dimensions win)))
+           (new-wid (inexact->exact (floor (* amount orig-wid))))
+           (new-hgt (inexact->exact (floor (* amount orig-hgt)))))
+      ;; this expects integers ("800 600") and fails on floats ("800. 600.")
+      (resize-window-with-hints* win new-wid new-hgt)))
+
+  (define (double-window-size w)
+    (resize-by-factor w 2))
+
+  (define (halve-window-size w)
+    (resize-by-factor w 0.5))
+
 ;;; hook functions
 
   (define (lost-window w)
@@ -490,33 +541,51 @@
   (add-hook 'unmap-notify-hook lost-window t)
   (add-hook 'destroy-notify-hook lost-window t)
 
-
 ;;; Entry points
 
   (define (move-window-interactively w)
-    "Move the window interactively using the mouse."
+    "Move the window interactively, with mouse or keyboard."
     (do-move-resize w 'move))
 
   (define (resize-window-interactively w)
-    "Resize the window interactively using the mouse."
+    "Resize the window interactively, with mouse or keyboard."
     (do-move-resize w 'resize))
 
   (define (move-selected-window)
-    "Wait for the user to select a window, then interactively move that window."
+    "Wait for the user to select a window, then interactively move
+that window."
     (let ((w (select-window)))
       (when w
 	(move-window-interactively w))))
 
   (define (resize-selected-window)
-    "Wait for the user to select a window, then interactively resize that window."
+    "Wait for the user to select a window, then interactively resize
+that window."
     (let ((w (select-window)))
       (when w
 	(resize-window-interactively w))))
 
+  ;; Move Window To Center
+
+  (define (move-window-center w)
+    (move-window-to w
+                    (quotient (- (screen-width)
+                                 (car (window-frame-dimensions w))) 2)
+                    (quotient (- (screen-height)
+                                 (cdr (window-frame-dimensions w))) 2)))
+
   ;;###autoload
+  (define-command 'move-window-center
+    move-window-center #:spec "%W")
   (define-command 'move-window-interactively
     move-window-interactively #:spec "%W")
   (define-command 'resize-window-interactively
     resize-window-interactively #:spec "%W")
-  (define-command 'move-selected-window move-selected-window)
-  (define-command 'resize-selected-window resize-selected-window))
+  (define-command 'move-selected-window
+    move-selected-window)
+  (define-command 'resize-selected-window
+    resize-selected-window)
+  (define-command 'double-window-size
+    double-window-size #:spec "%W")
+  (define-command 'halve-window-size
+    halve-window-size #:spec "%W"))
