@@ -1,30 +1,28 @@
 ;; frames.jl -- handle window framing
-;; $Id: frames.jl,v 1.81 2001/01/20 02:40:28 jsh Exp $
 
 ;; Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
-;; This file is part of sawmill.
+;; This file is part of sawfish.
 
-;; sawmill is free software; you can redistribute it and/or modify it
+;; sawfish is free software; you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; sawmill is distributed in the hope that it will be useful, but
+;; sawfish is distributed in the hope that it will be useful, but
 ;; WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with sawmill; see the file COPYING.  If not, write to
+;; along with sawfish; see the file COPYING.  If not, write to
 ;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
 (define-structure sawfish.wm.frames
 
     (compound-interface
      (structure-interface sawfish.wm.frames.subrs)
-     (export custom-set-frame-style
-	     after-setting-frame-option
+     (export after-setting-frame-option
 	     define-frame-type-mapper
 	     add-frame-style
 	     reload-frame-style
@@ -33,10 +31,14 @@
 	     rebuild-frames-with-style
 	     reframe-windows-with-style
 	     set-frame-style
+	     apply-frame-style
+	     apply-frame-style-and-save
 	     mark-frame-style-editable
 	     frame-style-editable-p
 	     window-type
 	     set-window-type
+	     push-window-type
+	     pop-window-type
 	     window-type-remove-title
 	     window-type-remove-border
 	     window-type-add-title
@@ -48,16 +50,18 @@
 	     add-frame-class
 	     frame-class-removed-p
 	     set-frame-part-value
+	     remove-frame-part-value
 	     def-frame-class
-	     define-frame-class))
-	    
+	     define-frame-class
+	     update-frame-font-color))
+
     (open rep
 	  rep.system
 	  rep.regexp
 	  rep.io.files
 	  rep.io.timers
 	  sawfish.wm.frames.subrs
-	  sawfish.wm.windows.subrs
+	  sawfish.wm.windows
 	  sawfish.wm.misc
 	  sawfish.wm.custom
 	  sawfish.wm.commands
@@ -84,6 +88,9 @@
   ;;	unframed		no frame at all
   ;;	icon
   ;;	dock
+  ;;	menu, toolbar
+  ;;	utility
+  ;;	splash
 
   ;; There is also a similar concept of frame types. The window type
   ;; never changes (unless the user explicitly does so). But the frame
@@ -106,7 +113,6 @@
   ;; non-maximized. (In fact the frame-type != window-type scheme is
   ;; largely just for compatibility)
 
-
 ;;; custom support
 
   (define (custom-make-frame-style-widget)
@@ -114,33 +120,9 @@
 
   (put 'frame-style 'custom-widget custom-make-frame-style-widget)
 
-
 ;;; variables etc
 
-  (defvar frame-part-classes
-    '((title . ((cursor . hand2)
-		(keymap . title-keymap)))
-      (menu-button . ((keymap . menu-button-keymap)))
-      (close-button . ((keymap . close-button-keymap)))
-      (iconify-button . ((keymap . iconify-button-keymap)))
-      (maximize-button . ((keymap . maximize-button-keymap)))
-      (shade-button . ((keymap . shade-button-keymap)))
-      (top-border . ((cursor . top_side)
-		     (keymap . border-keymap)))
-      (left-border . ((cursor . left_side)
-		      (keymap . border-keymap)))
-      (right-border . ((cursor . right_side)
-		       (keymap . border-keymap)))
-      (bottom-border . ((cursor . bottom_side)
-			(keymap . border-keymap)))
-      (top-left-corner . ((cursor . top_left_corner)
-			  (keymap . border-keymap)))
-      (top-right-corner . ((cursor . top_right_corner)
-			   (keymap . border-keymap)))
-      (bottom-left-corner . ((cursor . bottom_left_corner)
-			     (keymap . border-keymap)))
-      (bottom-right-corner . ((cursor . bottom_right_corner)
-			      (keymap . border-keymap))))
+  (defvar frame-part-classes nil
     "Alist of (CLASS . ALIST) associating classes of frame parts with state
 they inherit.")
 
@@ -155,35 +137,47 @@ that overrides settings set elsewhere.")
     "Default frame style:"
     :type frame-style
     :widget-flags (expand-vertically)
-    :user-level novice
     :group appearance
     :after-set (lambda () (after-setting-default-frame)))
 
   (defcustom reload-themes-when-changed t
     "Automatically reload themes when they are updated."
     :type boolean
-    :user-level expert
-    :group misc)
+    :group appearance)
 
-  (defcustom frame-type-fallback-alist
+  (defvar frame-type-fallback-alist
     '((transient . default)
       (shaped . default)
       (shaped-transient . shaped)
       (shaded . shaped)
       (shaded-transient . shaped-transient)
       (icon . shaped-transient)
-      (dock . icon))
-    "Frame type fallbacks:"
-    :tooltip "Associate frame types with type to try if the theme doesn't \
-implement the requested type."
-    :type (alist ((symbol default shaped transient
-			  shaped-transient icon doc) "From")
-		 ((symbol default shaped transient
-			  shaped-transient icon doc) "To"))
-    :widget-flags (expand-vertically framed)
-    :group appearance
-    :user-level expert
-    :after-set (lambda () (after-setting-frame-option)))
+      (dock . unframed)
+      (utility . default)
+      (shaded-utility . shaded-transient)
+      (toolbar . unframed)
+      (menu . unframed)
+      (splash . unframed)
+      ;; (shaded-toolbar . shaded-utility)
+      ;; (shaded-menu . shaded-utility)
+      )
+    "Frame type fallbacks. This is an alist whose elements are
+(required-type . fallback-type). When REQUIRED-TYPE is not supported
+by the current theme, then FALLBACK-TYPE is used instead.")
+
+  ;; Re-Coloring Options for Themes
+
+  (define (update-frame-font-color)
+    (if use-custom-font-color
+	(mapc (lambda (fc)
+		(set-frame-part-value fc 'foreground
+                                      (list frame-font-inactive-color
+                                            frame-font-active-color)
+                                      t))
+              '(title tab))
+      (mapc (lambda (fc) (remove-frame-part-value fc 'foreground t))
+            '(title tab)))
+    (mapc rebuild-frame (managed-windows)))
 
   (defvar theme-update-interval 60
     "Number of seconds between checking if theme files have been modified.")
@@ -205,7 +199,7 @@ implement the requested type."
     "List of directories from which themes may be loaded.")
 
   (define frame-styles nil
-    "List of (NAME . FUNCTION) defining all loaded frame styles.")
+          "List of (NAME . FUNCTION) defining all loaded frame styles.")
 
   ;; List of (NAME FILENAME MODTIME) mapping loaded frame styles to the
   ;; files they were loaded from; used to check if the theme needs reloading
@@ -215,21 +209,23 @@ implement the requested type."
   (define editable-frame-styles nil)
 
   (define frame-type-mappers '()
-    "List of functions that map (WINDOW FRAME-TYPE) -> FRAME-TYPE. Used when
-deciding which frame type to ask a theme to generate.")
+          "List of functions that map (WINDOW FRAME-TYPE) ->
+FRAME-TYPE. Used when deciding which frame type to ask a theme to
+generate.")
 
   ;; list of (REGEXP DIR-EXPAND NAME-EXPAND)
   (defvar theme-suffix-regexps
-    '(("^(.*)/(.*)\\.tar(\\.gz|\\.Z|\\.bz2|)$" "\\0#tar/\\2" "\\2")))
+    '(("^(.*)/(.*)\\.tar(\\.gz|\\.Z|\\.bz2|\\.xz|\\.lzma|)$"
+       "\\0#tar/\\2" "\\2")))
 
-  (defvar theme-suffixes '("" ".tar" ".tar.gz" ".tar.Z" ".tar.bz2"))
+  (defvar theme-suffixes '("" ".tar" ".tar.gz" ".tar.Z" ".tar.bz2"
+                           ".tar.xz" ".tar.lzma"))
 
   (defvar themes-are-gaolled t
     "When non-nil themes are assumed to be malicious.")
 
   (defvar sawfish-themer-program "sawfish-themer")
 
-
 ;;; defcustom's for some built-in variables
 
   (defcustom default-font nil
@@ -237,17 +233,32 @@ deciding which frame type to ask a theme to generate.")
     :group appearance
     :type font
     :widget-flags (expand-horizontally)
-    :user-level novice
     :after-set (lambda () (after-setting-frame-option)))
 
-  (defcustom default-bevel-percent nil
-    "Bevel intensity: \\wpercent."
+  (defvar default-bevel-percent nil
+    "Bevel intensity as a percentage.")
+
+  ;; frame fonts & colors
+  (defcustom use-custom-font-color '()
+    "Use custom font colors for frames"
+    :type boolean
     :group appearance
-    :type (number 0 100)
-    :user-level expert
-    :after-set (lambda () (after-setting-frame-option)))
+    :after-set (lambda () (update-frame-font-color)))
+
+  (defcustom frame-font-active-color "black"
+    "Font color for active frames"
+    :type color
+    :group appearance
+    :depends use-custom-font-color
+    :after-set (lambda () (update-frame-font-color)))
+
+  (defcustom frame-font-inactive-color "black"
+    "Font color for inactive frames"
+    :type color
+    :group appearance
+    :depends use-custom-font-color
+    :after-set (lambda () (update-frame-font-color)))
 
-
 ;;; managing frame types
 
   (define (define-frame-type-mapper fun)
@@ -258,23 +269,24 @@ deciding which frame type to ask a theme to generate.")
     ;; 1. map window type to actual frame type
     (let loop-1 ((rest frame-type-mappers)
 		 (type (window-type w)))
-      (if (null rest)
-	  ;; found the final frame type, so,
-	  ;; 2. find the closest type that the style implements to this
-	  (let loop-2 ((type type)
-		       (seen (list type)))
-	    (cond ((eq type 'unframed) nil-frame)
-		  ((style w type))
-		  (t (let ((next (or (cdr (assq type frame-type-fallback-alist))
-				     'unframed)))
-		       (if (memq next seen)
-			   ;; been here before..
-			   nil-frame
-			 (loop-2 next (cons next seen)))))))
-	;; else, apply this transformation and keep looping
-	(loop-1 (cdr rest) ((car rest) w type)))))
+         (if (null rest)
+             ;; found the final frame type, so,
+             ;; 2. find the closest type that the style implements to this
+             (let loop-2 ((type type)
+                          (seen (list type)))
+                  (cond ((eq type 'unframed) nil-frame)
+                        ((style w type))
+                        (t (let ((next (or
+                                        (cdr (assq type
+                                                   frame-type-fallback-alist))
+                                        'unframed)))
+                             (if (memq next seen)
+                                 ;; been here before..
+                                 nil-frame
+                               (loop-2 next (cons next seen)))))))
+           ;; else, apply this transformation and keep looping
+           (loop-1 (cdr rest) ((car rest) w type)))))
 
-
 ;;; managing frame styles
 
   (define (add-frame-style name function)
@@ -319,7 +331,6 @@ deciding which frame type to ask a theme to generate.")
 		  (reload-frame-style style))))
 	    frame-style-files)))
 
-
 ;;; applying frame styles to windows
 
   (define (reframe-window w)
@@ -360,7 +371,27 @@ deciding which frame type to ask a theme to generate.")
       (call-window-hook 'window-state-change-hook w (list '(frame-style)))
       (reframe-window w)))
 
-
+  ;; return true iff successful
+  (define (apply-frame-style style)
+    (let ((old-style default-frame-style))
+      (condition-case nil
+	  (progn
+	    (setq default-frame-style style)
+	    (after-setting-default-frame)
+	    t)
+	(error
+	 (setq default-frame-style old-style)
+	 (after-setting-frame-option)
+	 nil))))
+
+  (define (save-current-frame-style)
+    (require 'sawfish.wm.customize)
+    (customize-set 'default-frame-style default-frame-style))
+
+  (define (apply-frame-style-and-save style)
+    (when (apply-frame-style style)
+      (save-current-frame-style)))
+
 ;;; editable frame styles
 
   (define (mark-frame-style-editable style)
@@ -378,7 +409,6 @@ deciding which frame type to ask a theme to generate.")
 	    (system (format nil "%s %s &" sawfish-themer-program dir))))))
     #:spec (lambda () (list default-frame-style)))
 
-
 ;;; kludge different window decors by modifying the assumed window type
 
   (define (window-type w)
@@ -392,11 +422,34 @@ deciding which frame type to ask a theme to generate.")
 	    'default))))
 
   (define (set-window-type w type)
-    (unless (eq (window-get w 'type) type)
+    (if (window-get w 'type/saved)
+	(window-put w 'type/saved type)
+      (unless (eq (window-get w 'type) type)
+	(window-put w 'type type)
+	(call-window-hook 'window-state-change-hook w (list '(type)))
+	(reframe-window w))))
+
+  ;; XXX do something with KEY. (It's a unique symbol used to mark
+  ;; XXX the different users of these functions)
+
+  (define (push-window-type w type key)
+    (window-put w 'type/key key)
+    (unless (eq (window-get w 'type type))
+      (unless (window-get w 'type/saved)
+	(window-put w 'type/saved (window-get w 'type)))
       (window-put w 'type type)
       (call-window-hook 'window-state-change-hook w (list '(type)))
       (reframe-window w)))
 
+  (define (pop-window-type w key)
+    (when (and (window-get w 'type/saved)
+	       (eq (window-get w 'type/key) key))
+      (window-put w 'type (window-get w 'type/saved))
+      (window-put w 'type/saved nil)
+      (window-put w 'type/key nil))
+    (call-window-hook 'window-state-change-hook w (list '(type)))
+    (reframe-window w))
+
   (define (window-type-remove-title type)
     (case type
       ((default) 'transient)
@@ -427,7 +480,6 @@ deciding which frame type to ask a theme to generate.")
 	    (lambda (w) (set-window-type w type)) #:spec "%W"))
 	'(default transient shaped shaped-transient unframed))
 
-
 ;;; loading ``themes'' (currently just frame styles)
 
   (define (frame-style-directory dir #!optional get-name)
@@ -473,7 +525,7 @@ deciding which frame type to ask a theme to generate.")
 		(gaol-load (expand-file-name "theme.jl" dir) gaol)
 		(define-gaol-structure (intern (concat "themes."
 						       (symbol-name name)))
-				       gaol))
+                  gaol))
 	    (load (expand-file-name "theme" dir) nil t))))))
 
   (define (find-all-frame-styles #!optional sorted)
@@ -524,7 +576,6 @@ deciding which frame type to ask a theme to generate.")
        (check . ,(eq (window-type w) 'unframed))
        (group . menu-type))))
 
-
 ;;; removing frame parts
 
   (define (remove-frame-class w class)
@@ -542,7 +593,6 @@ deciding which frame type to ask a theme to generate.")
   (define (frame-class-removed-p w class)
     (memq class (window-get w 'removed-classes)))
 
-
 ;;; manipulating the frame part classes variables
 
   (define (set-frame-part-value class key value #!optional override)
@@ -556,10 +606,17 @@ deciding which frame type to ask a theme to generate.")
 	(set var (cons (cons class (list (cons key value)))
 		       (symbol-value var))))))
 
+  (define (remove-frame-part-value class key #!optional override)
+    (let* ((fpcs (if override override-frame-part-classes frame-part-classes))
+           (item (assq class fpcs)))
+      (when item
+        (setcdr item (delete-if (lambda (it) (eq (car it) key))
+                                (cdr item))))))
+
   ;; (def-frame-class shade-button '((cursor . foo) ...)
   ;;   (bind-keys shade-button-keymap
   ;;     "Button1-Off" 'toggle-window-shaded))
-  ;; 
+  ;;
   ;; the idea being that it will only create the frame part if it doesn't
   ;; exist, it will add all properties from the second argument unless
   ;; they're already set, then create and initialise the keymap from the
@@ -594,9 +651,44 @@ deciding which frame type to ask a theme to generate.")
 	  (set-frame-part-value class 'keymap map-name)))
       ok-to-bind))
 
-
+  (define ((cursor-for-frame-part part) w)
+    (require 'sawfish.wm.state.maximize)
+    (if (frame-part-movable-p w part)
+	(case part
+	  ((top-border) 'top_side)
+	  ((bottom-border) 'bottom_side)
+	  ((left-border) 'left_side)
+	  ((right-border) 'right_side)
+	  ((top-left-corner) 'top_left_corner)
+	  ((top-right-corner) 'top_right_corner)
+	  ((bottom-left-corner) 'bottom_left_corner)
+	  ((bottom-right-corner) 'bottom_right_corner))
+      nil))
+
 ;;; initialisation
 
+  (define-frame-class 'menu-button '((keymap . menu-button-keymap)))
+  (define-frame-class 'close-button '((keymap . close-button-keymap)))
+  (define-frame-class 'iconify-button '((keymap . iconify-button-keymap)))
+  (define-frame-class 'maximize-button '((keymap . maximize-button-keymap)))
+  (define-frame-class 'shade-button '((keymap . shade-button-keymap)))
+  (define-frame-class 'sticky-button '((keymap . sticky-button-keymap)))
+  (define-frame-class 'lock-button '((keymap . lock-button-keymap)))
+  (define-frame-class 'rename-button '((keymap . rename-button-keymap)))
+  (define-frame-class 'move-resize-button '((keymap . move-resize-button-keymap)))
+  (define-frame-class 'raise-lower-button '((keymap . raise-lower-button-keymap)))
+
+  (define-frame-class 'title `((keymap . title-keymap)
+			       (cursor . ,(cursor-for-frame-part 'title))))
+
+  (mapc (lambda (x)
+	  (define-frame-class x `((cursor . ,(cursor-for-frame-part x))
+				  (keymap . border-keymap))))
+	'(top-border bottom-border
+                     left-border right-border
+                     top-left-corner top-right-corner
+                     bottom-left-corner bottom-right-corner))
+
   (gaol-add add-frame-style reframe-window rebuild-frames-with-style
 	    reframe-windows-with-style reframe-all-windows window-type
 	    def-frame-class define-frame-class after-setting-frame-option
@@ -610,5 +702,14 @@ deciding which frame type to ask a theme to generate.")
 
   (make-timer frames-on-idle theme-update-interval)
 
-  (sm-add-saved-properties 'type 'ignored 'frame-style)
-  (add-swapped-properties 'frame-active-color 'frame-inactive-color))
+  (add-hook 'sm-window-save-functions
+	    (lambda (w)
+	      (cond ((window-get w 'type/saved)
+		     (list (cons 'type (window-get w 'type/saved))))
+		    ((window-get w 'type)
+		     (list (cons 'type (window-get w 'type)))))))
+
+  (sm-add-saved-properties 'ignored 'frame-style)
+  (sm-add-restored-properties 'type)
+  (add-swapped-properties 'frame-active-color 'frame-inactive-color)
+  )
