@@ -1,28 +1,29 @@
 ;; stacking.jl -- window stacking
-;; $Id: stacking.jl,v 1.59 2002/10/18 06:46:16 jsh Exp $
 
 ;; Copyright (C) 1999 John Harper <john@dcs.warwick.ac.uk>
 
-;; This file is part of sawmill.
+;; This file is part of sawfish.
 
-;; sawmill is free software; you can redistribute it and/or modify it
+;; sawfish is free software; you can redistribute it and/or modify it
 ;; under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; sawmill is distributed in the hope that it will be useful, but
+;; sawfish is distributed in the hope that it will be useful, but
 ;; WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with sawmill; see the file COPYING.  If not, write to
+;; along with sawfish; see the file COPYING.  If not, write to
 ;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
 (define-structure sawfish.wm.stacking
 
     (export save-stacking-order
 	    mapped-stacking-order
+            window-obscured
+            stacking-visibility
 	    raise-window
 	    lower-window
 	    stack-window-above
@@ -46,7 +47,8 @@
 	  sawfish.wm.custom
 	  sawfish.wm.session.init
 	  sawfish.wm.workspace
-	  sawfish.wm.state.transient)
+          sawfish.wm.state.transient
+          sawfish.wm.util.rects)
 
   ;; Each window will have a `depth' property--an integer, zero
   ;; represents the level of normal windows, negative for windows below
@@ -70,7 +72,6 @@
   ;; this will always return an integer
   (define (window-depth w) (window-get w 'depth))
 
-
 ;;; constraint mechanics (predicates actually..)
 
   (define (stacking-constraint:layer w)
@@ -93,20 +94,20 @@
 				   (> (window-depth x) w-depth)) parents))
 	(setq children (delete-if (lambda (x)
 				    (< (window-depth x) w-depth)) children)))
-				 
+
       (lambda (above below)
 	(and (or (null parents)
 		 ;; All parents must be below W
 		 (let loop ((rest parents))
-		   (cond ((null rest) t)
-			 ((memq (car rest) above) nil)
-			 (t (loop (cdr rest))))))
+                      (cond ((null rest) t)
+                            ((memq (car rest) above) nil)
+                            (t (loop (cdr rest))))))
 	     (or (null children)
 		 ;; All children must be above W
 		 (let loop ((rest children))
-		   (cond ((null rest) t)
-			 ((memq (car rest) below) nil)
-			 (t (loop (cdr rest))))))))))
+                      (cond ((null rest) t)
+                            ((memq (car rest) below) nil)
+                            (t (loop (cdr rest))))))))))
 
   (define (stacking-constraint:transients-above-all w)
     (let ((w-depth (window-depth w)))
@@ -116,27 +117,27 @@
 	  (lambda (above below)
 	    (declare (unused below))
 	    (let loop ((rest above))
-	      (cond ((null rest) t)
-		    ((and (not (window-transient-p (car rest)))
-			  (<= (window-depth (car rest)) w-depth)) nil)
-		    (t (loop (cdr rest))))))
+                 (cond ((null rest) t)
+                       ((and (not (window-transient-p (car rest)))
+                             (<= (window-depth (car rest)) w-depth)) nil)
+                       (t (loop (cdr rest))))))
 	;; ensure no transients below W, unless they're depth < W
 	(lambda (above below)
 	  (declare (unused above))
 	  (let loop ((rest below))
-	    (cond ((null rest) t)
-		  ((and (window-transient-p (car rest))
-			(>= (window-depth (car rest)) w-depth)) nil)
-		  (t (loop (cdr rest)))))))))
+               (cond ((null rest) t)
+                     ((and (window-transient-p (car rest))
+                           (>= (window-depth (car rest)) w-depth)) nil)
+                     (t (loop (cdr rest)))))))))
 
   (define (combine-constraints constraints)
     "Combine the list of secondary constraint functions into a single
 function (using logical `and' combinator)."
     (lambda (above below)
       (let loop ((rest constraints))
-	(cond ((null rest) t)
-	      ((not ((car rest) above below)) nil)
-	      (t (loop (cdr rest)))))))
+           (cond ((null rest) t)
+                 ((not ((car rest) above below)) nil)
+                 (t (loop (cdr rest)))))))
 
   (defvar basic-stacking-constraints (list stacking-constraint:layer)
     "List of stacking constraint functions to adhere to (excluding transient
@@ -158,7 +159,6 @@ for a particular stacking configuration, it will return non-`nil'."
 		      basic-stacking-constraints))
 	       (t basic-stacking-constraints)))))
 
-
 ;;; utilities
 
   ;; returns (BEFORE . AFTER), modifies LST
@@ -170,14 +170,14 @@ reverse order), and the list of windows occurring in LST after PIVOT.
 LST is destructively modified by this procedure."
     (let loop ((rest lst)
 	       (before '()))
-      (cond ((null rest)
-	     (cons before rest))
-	    ((eq (car rest) pivot)
-	     (cons before (cdr rest)))
-	    (t
-	     (let ((next (cdr rest)))
-	       (rplacd rest before)
-	       (loop next rest))))))
+         (cond ((null rest)
+                (cons before rest))
+               ((eq (car rest) pivot)
+                (cons before (cdr rest)))
+               (t
+                (let ((next (cdr rest)))
+                  (rplacd rest before)
+                  (loop next rest))))))
 
   (define (stack-rotate-upwards cell)
     "Given a cons cell containing two lists of windows `(ABOVE . BELOW)',
@@ -208,7 +208,54 @@ the empty list."
   (define (mapped-stacking-order)
     (delete-if-not window-mapped-p (stacking-order)))
 
-
+  (define (window-obscured window)
+    "Check whether WINDOW is obscured.  Return `t' if WINDOW is fully obscured
+by some other window, otherwise a list of windows partially obscuring WINDOW.
+In particular return `nil' if WINDOW is unobscured.  Note that if a list of
+partially obscuring windows is returned, taken together they may or may not
+fully obscure WINDOW."
+    (define w window)
+    (define ws (nearest-workspace-with-window w current-workspace))
+    (let loop ((stack (stacking-order))
+               (obs nil))
+         (if (null stack)               ; Should not happen
+             obs
+           (let ((w2 (car stack)))
+             (cond ((eq w2 w) obs)
+                   ((and (window-visible-p w2)
+                         (window-appears-in-workspace-p w2 ws))
+                    (case (apply rect-obscured
+                                 (rectangles-from-windows (list w w2)))
+                      ((unobscured) (loop (cdr stack) obs))
+                      ((fully-obscured) t)
+                      (t (loop (cdr stack) (cons w2 obs))))) ; Partially
+                   (t (loop (cdr stack) obs)))))))
+
+  (define (stacking-visibility window)
+    "Compute the visibility of WINDOW from the stacking order.  This should
+work even with the Composite extension, which appears to disable
+VisibilityNotify events.  Note that deciding between fully and partially
+obscured may require quite a bit of computation.  If you do not need that
+distinction, window-obscured should be faster."
+    (define (rect-list-minus rs s tail)
+      (if (null rs)
+          tail
+        (rect-list-minus (cdr rs) s
+                         (rect-minus (car rs) s tail))))
+    (let ((obs (window-obscured window)))
+      (case obs
+        ((t) 'fully-obscured)
+        ((()) 'unobscured)
+        (t
+         (do ((unobs (rectangles-from-windows (list window))
+                     (rect-list-minus unobs (car robs) nil))
+              (robs (rectangles-from-windows obs) (cdr robs)))
+             ((or (null unobs) (null robs))
+              (if (null unobs)
+                  'fully-obscured
+                'partially-obscured)))))))
+
+
 ;;; stacking functions
 
   (define (raise-window w)
@@ -217,16 +264,16 @@ the empty list."
     (let ((constraint (make-constraint w))
 	  (stack (cons '() (delq w (stacking-order)))))
       (let loop ()
-	(cond ((constraint (car stack) (cdr stack))
-	       (if (car stack)
-		   (x-lower-window w (car (car stack)))
-		 (x-raise-window w (car (cdr stack)))))
-	      ((null (cdr stack))
-	       ;; no position
-	       nil)
-	      (t
-	       (stack-rotate-downwards stack)
-	       (loop))))))
+           (cond ((constraint (car stack) (cdr stack))
+                  (if (car stack)
+                      (x-lower-window w (car (car stack)))
+                    (x-raise-window w (car (cdr stack)))))
+                 ((null (cdr stack))
+                  ;; no position
+                  nil)
+                 (t
+                  (stack-rotate-downwards stack)
+                  (loop))))))
 
   (define (lower-window w)
     "Lower the window to its lowest allowed position in the stacking order."
@@ -234,17 +281,17 @@ the empty list."
 	  (stack (cons (nreverse (delq w (stacking-order))) '())))
       ;; work upwards from bottom
       (let loop ()
-	(cond ((constraint (car stack) (cdr stack))
-	       ;; found the lowest position
-	       (if (cdr stack)
-		   (x-raise-window w (car (cdr stack)))
-		 (x-lower-window w (car (car stack)))))
-	      ((null (car stack))
-	       ;; no possible position..
-	       nil)
-	      (t
-	       (stack-rotate-upwards stack)
-	       (loop))))))
+           (cond ((constraint (car stack) (cdr stack))
+                  ;; found the lowest position
+                  (if (cdr stack)
+                      (x-raise-window w (car (cdr stack)))
+                    (x-lower-window w (car (car stack)))))
+                 ((null (car stack))
+                  ;; no possible position..
+                  nil)
+                 (t
+                  (stack-rotate-upwards stack)
+                  (loop))))))
 
   (define (stack-window-above above below)
     "Change the stacking of window ABOVE so that it is as closely above window
@@ -254,17 +301,17 @@ BELOW as possible."
 		  (delq above (stacking-order)) below)))
       (rplacd stack (cons below (cdr stack)))
       (let loop ()
-	(cond ((constraint (car stack) (cdr stack))
-	       ;; found a suitable position
-	       (if (car stack)
-		   (x-lower-window above (car (car stack)))
-		 (x-raise-window above (car (cdr stack)))))
-	      ((null (car stack))
-	       ;; reached the top
-	       nil)
-	      (t
-	       (stack-rotate-upwards stack)
-	       (loop))))))
+           (cond ((constraint (car stack) (cdr stack))
+                  ;; found a suitable position
+                  (if (car stack)
+                      (x-lower-window above (car (car stack)))
+                    (x-raise-window above (car (cdr stack)))))
+                 ((null (car stack))
+                  ;; reached the top
+                  nil)
+                 (t
+                  (stack-rotate-upwards stack)
+                  (loop))))))
 
   (define (stack-window-below below above)
     "Change the stacking of window BELOW so that it is as closely below window
@@ -274,16 +321,16 @@ ABOVE as possible."
 		  (delq below (stacking-order)) above)))
       (rplaca stack (cons above (car stack)))
       (let loop ()
-	(cond ((constraint (car stack) (cdr stack))
-	       (if (cdr stack)
-		   (x-raise-window below (car (cdr stack)))
-		 (x-lower-window below (car (car stack)))))
-	      ((null (cdr stack))
-	       ;; reached the bottom
-	       nil)
-	      (t
-	       (stack-rotate-downwards stack)
-	       (loop))))))
+           (cond ((constraint (car stack) (cdr stack))
+                  (if (cdr stack)
+                      (x-raise-window below (car (cdr stack)))
+                    (x-lower-window below (car (car stack)))))
+                 ((null (cdr stack))
+                  ;; reached the bottom
+                  nil)
+                 (t
+                  (stack-rotate-downwards stack)
+                  (loop))))))
 
   (define (restack-window w)
     "Assuming that the current stacking order is in a consistent state
@@ -318,7 +365,8 @@ order they are stacked within the layer (top to bottom)."
       (call-window-hook 'window-state-change-hook w (list '(stacking)))))
 
   (define (window-on-top-p w)
-    "Return t if window W is as high as it can legally go in the stacking order."
+    "Return t if window W is as high as it can legally go in the
+stacking order."
     (let* ((constraint (make-constraint w))
 	   (order (delete-if-not
 		   ;; XXX doesn't handle viewports..
@@ -330,17 +378,17 @@ order they are stacked within the layer (top to bottom)."
 	   (old-posn (- (length order) (length (memq w order))))
 	   (stack (cons '() (delq w order))))
       (let loop ()
-	(if (or (constraint (car stack) (cdr stack))
-		(null (cdr stack)))
-	    ;; found highest position
-	    (= (length (car stack)) old-posn)
-	  (stack-rotate-downwards stack)
-	  (loop)))))
+           (if (or (constraint (car stack) (cdr stack))
+                   (null (cdr stack)))
+               ;; found highest position
+               (= (length (car stack)) old-posn)
+             (stack-rotate-downwards stack)
+             (loop)))))
 
   (define (raise-lower-window w)
     "If the window is at its highest possible position, then lower it to its
 lowest possible position. Otherwise raise it as far as allowed."
-    (if (or (eq (window-visibility w) 'unobscured)
+    (if (or (not (window-obscured w))
 	    (window-on-top-p w))
 	(lower-window w)
       (raise-window w)))
@@ -362,7 +410,6 @@ lowest possible position. Otherwise raise it as far as allowed."
   (define-command 'raise-window-depth raise-window-depth #:spec "%W")
   (define-command 'lower-window-depth lower-window-depth #:spec "%W")
 
-
 ;;; stacking groups of windows
 
   ;; Helper function for raise-windows and lower-windows. Calls (FIRST
@@ -382,26 +429,26 @@ lowest possible position. Otherwise raise it as far as allowed."
     ;; must be a valid stacking configuration (highest-to-lowest).  Do
     ;; the restacking in such a way that no part of any window in
     ;; ORDER is exposed during the restacking that is not exposed
-    ;; afterwards. 
+    ;; afterwards.
 
     (define (highest-window order)
-      ;; find the first window in ORDER that is allowed 
+      ;; find the first window in ORDER that is allowed
       ;; to be above all other windows in ORDER...
       (let loop ((rest order))
-	(cond ((null rest) nil)
-	      (((make-constraint (car rest)) '() (remq (car rest) order))
-	       (car rest))
-	      (t (loop (cdr rest))))))
+           (cond ((null rest) nil)
+                 (((make-constraint (car rest)) '() (remq (car rest) order))
+                  (car rest))
+                 (t (loop (cdr rest))))))
 
     ;; Cons up the new order (in reverse), by picking the most-raisable
     ;; each time until no windows are left, then commit that ordering
     (let loop ((rest (cons w (remq w order)))
 	       (out '()))
-      (if (null rest)
-	  (apply-group-order (nreverse out) raise-window stack-window-below)
-	(let ((highest (or (highest-window rest)
-			   (error "Stacking constraint failed"))))
-	  (loop (delq highest rest) (cons highest out))))))
+         (if (null rest)
+             (apply-group-order (nreverse out) raise-window stack-window-below)
+           (let ((highest (or (highest-window rest)
+                              (error "Stacking constraint failed"))))
+             (loop (delq highest rest) (cons highest out))))))
 
   (define (lower-windows w order)
 
@@ -410,32 +457,31 @@ lowest possible position. Otherwise raise it as far as allowed."
       ;; windows in ORDER
       (let ((reversed (reverse order)))
 	(let loop ((rest reversed))
-	  (cond ((null rest) nil)
-		(((make-constraint (car rest))
-		  (remq (car rest) reversed) '()) (car rest))
-		(t (loop (cdr rest)))))))
+             (cond ((null rest) nil)
+                   (((make-constraint (car rest))
+                     (remq (car rest) reversed) '()) (car rest))
+                   (t (loop (cdr rest)))))))
 
     (let loop ((rest (nconc (remq w order) (list w)))
 	       (out '()))
-      (if (null rest)
-	  (apply-group-order (nreverse out) lower-window stack-window-above)
-	(let ((lowest (or (lowest-window rest)
-			  (error "Stacking constraint failed"))))
-	  (loop (delq lowest rest) (cons lowest out))))))
+         (if (null rest)
+             (apply-group-order (nreverse out) lower-window stack-window-above)
+           (let ((lowest (or (lowest-window rest)
+                             (error "Stacking constraint failed"))))
+             (loop (delq lowest rest) (cons lowest out))))))
 
   (define (raise-lower-windows w order)
-    (if (or (eq (window-visibility w) 'unobscured)
+    (if (or (not (window-obscured w))
 	    (and (window-on-top-p (car order))
 		 ;; look for the group as a block.. this is a heuristic
 		 (let loop ((rest (memq (car order) (stacking-order))))
-		   (cond ((null rest) nil)
-			 ((eq (car rest) w) t)
-			 ((memq (car rest) order) (loop (cdr rest)))
-			 (t nil)))))
+                      (cond ((null rest) nil)
+                            ((eq (car rest) w) t)
+                            ((memq (car rest) order) (loop (cdr rest)))
+                            (t nil)))))
 	(lower-windows w order)
       (raise-windows w order)))
 
-
 ;;; hooks
 
   ;; Called from the add-window-hook
