@@ -1,25 +1,22 @@
-#| windows.jl -- miscellaneous window mgmt functions
-
-   $Id: windows.jl,v 1.31 2002/11/14 06:58:00 jsh Exp $
-
-   Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
-
-   This file is part of librep.
-
-   librep is free software; you can redistribute it and/or modify it
-   under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2, or (at your option)
-   any later version.
-
-   librep is distributed in the hope that it will be useful, but
-   WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
-
-   You should have received a copy of the GNU General Public License
-   along with Jade; see the file COPYING.  If not, write to
-   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
-|#
+;; windows.jl -- miscellaneous window mgmt functions
+;;
+;; Copyright (C) 2000 John Harper <john@dcs.warwick.ac.uk>
+;;
+;; This file is part of sawfish.
+;;
+;; sawfish is free software; you can redistribute it and/or modify it
+;; under the terms of the GNU General Public License as published by
+;; the Free Software Foundation; either version 2, or (at your option)
+;; any later version.
+;;
+;; sawfish is distributed in the hope that it will be useful, but
+;; WITHOUT ANY WARRANTY; without even the implied warranty of
+;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+;; GNU General Public License for more details.
+;;
+;; You should have received a copy of the GNU General Public License
+;; along with sawfish; see the file COPYING.  If not, write to
+;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
 (define-structure sawfish.wm.windows
 
@@ -27,6 +24,8 @@
      (structure-interface sawfish.wm.windows.subrs)
      (export get-window-by-name
 	     get-window-by-name-re
+	     get-window-by-class
+	     get-window-by-class-re
 	     window-really-wants-input-p
 	     window-transient-p
 	     mark-window-as-transient
@@ -39,6 +38,7 @@
 	     warp-cursor-to-window
 	     activate-window
 	     constrain-dimension-to-hints
+	     constrain-aspect-to-hints
 	     resize-window-with-hints
 	     resize-window-with-hints*
 	     window-gravity
@@ -55,7 +55,9 @@
 	     window-avoided-p
 	     avoided-windows
 	     call-after-property-changed
-	     call-after-state-changed))
+	     call-after-state-changed
+	     rename-window-func
+	     rename-window-interactive))
 
     (open rep
 	  rep.system
@@ -67,34 +69,48 @@
 	  sawfish.wm.misc
 	  sawfish.wm.commands)
 
-  (defvar ignore-window-input-hint nil
-    "Give focus to windows even when they haven't asked for it.")
 
-  (defvar warp-to-window-offset (cons -1 -1)
-    "Offset (%) from window edges when warping pointer. A negative number
-means outside the left window edge.")
+  (defcustom ignore-window-input-hint nil
+    "Give focus to windows even when they haven't asked for it."
+    :type boolean
+    :group focus)
 
-  (defvar warp-to-window-enabled nil
-    "When false, disable warping the cursor to windows.")
- 
-  (defvar dont-avoid-ignored t
-    "When non-nil, ignored windows aren't avoided by default.")
+  (defgroup warp "Warping" :group misc)
 
-  (defvar avoid-by-default nil
-    "When non-nil, any unspecified windows are avoided by default.")
+  (defcustom warp-to-window-offset (cons -1 -1)
+    "Offset (%) from window edges when warping pointer."
+    :type (pair (number 1) (number 1))
+    :group (misc warp))
 
-  (defvar uniquify-name-format "%s [%d]"
-    "Format to create unique window names.")
+  (defcustom warp-to-window-enabled nil
+    "Whether to enable warping the cursor to windows."
+    :type boolean
+    :group (misc warp))
+
+  (defcustom dont-avoid-ignored t
+    "Whether to not avoid ignored windows by default."
+    :type boolean
+    :group misc)
+
+  (defcustom avoid-by-default nil
+    "Whether to avoid any unspecified window by default."
+    :type boolean
+    :group misc)
+
+  (defcustom uniquify-name-format "%s [%d]"
+    "Format to create unique window names."
+    :type string
+    :group misc)
 
   (defvar dock-window-properties
     '(window-list-skip cycle-skip fixed-position focus-click-through
-      avoid no-history never-iconify never-maximize sticky
-      sticky-viewport placed)
-    "List of properties set (to true) on windows marked as docks.")
+		       avoid no-history never-iconify never-maximize sticky
+		       sticky-viewport placed)
+    "List of properties set (to true) on windows when they are marked as docks.")
 
   (defvar desktop-window-properties
     '(fixed-position sticky sticky-viewport)
-    "List of properties set (to true) on windows marked as desktops.")
+    "List of properties set (to true) on windows when they are marked as desktops.")
 
   (defvar desktop-window-depth -4
     "The stacking depth of desktop windows.")
@@ -102,7 +118,6 @@ means outside the left window edge.")
   (defvar dock-window-depth 0
     "The stacking depth of dock windows.")
 
-
 ;;; finding windows, reading properties
 
   (define (get-window-by-name name)
@@ -117,6 +132,18 @@ Returns nil if no such window is found."
     (car (filter-windows (lambda (w)
 			   (string-match name (window-name w))))))
 
+  (define (get-window-by-class class)
+    "Find a window object whose window-class is CLASS. Returns nil if no such
+window is found."
+    (car (filter-windows (lambda (w)
+			   (string= (window-class w) class)))))
+
+  (define (get-window-by-class-re class)
+    "Find a window object whose window-class matches the regexp CLASS.
+Returns nil if no such window is found."
+    (car (filter-windows (lambda (w)
+			   (string-match class (window-class w))))))
+
   (define (window-really-wants-input-p w)
     "Return nil if window W should never be focused."
     (and (not (window-get w 'never-focus))
@@ -164,7 +191,7 @@ then be the numeric id of its parent window."
   (define-command 'focus-desktop focus-desktop)
 
   (define (dock-window-p arg)
-    "Return true if ARG represents a dock window (i.e. the GNOME panel)."
+    "Return true if ARG represents a dock window."
     (and (windowp arg) (window-get arg 'dock-type)))
 
   (define (mark-window-as-dock w)
@@ -174,6 +201,7 @@ then be the numeric id of its parent window."
     (window-put w 'dock-type t)
     (mapc (lambda (p)
 	    (window-put w p t)) dock-window-properties)
+    (set-window-type w 'dock)
     (set-window-depth w dock-window-depth))
 
   (define (window-in-cycle-p w #!key ignore-cycle-skip)
@@ -187,7 +215,7 @@ windows."
     "Return the class that window W belongs to, as a string. Returns `nil' if W
 has no associated class."
     (let ((prop (get-x-text-property w 'WM_CLASS)))
-      (and prop (aref prop 1))))
+      (and prop (> (length prop) 1) (aref prop 1))))
 
   (define (get-window-wm-protocols w)
     "Return a list of symbols defining the X11 window manager protocols
@@ -201,16 +229,15 @@ supported by client window W."
 	     (nreverse out))))))
 
   (define (window-supports-wm-protocol-p w atom)
-    "Return true if winow W includes ATOM in its `WM_PROTOCOLS' property."
+    "Return true if window W includes ATOM in its `WM_PROTOCOLS' property."
     (let* ((prop (get-x-property w 'WM_PROTOCOLS))
 	   (data (and prop (eq (car prop) 'ATOM) (nth 2 prop))))
       (when data
 	(let loop ((i 0))
-	  (cond ((= i (length data)) nil)
-		((eq (aref data i) atom) t)
-		(t (loop (1+ i))))))))
+             (cond ((= i (length data)) nil)
+                   ((eq (aref data i) atom) t)
+                   (t (loop (1+ i))))))))
 
-
 ;;; warping
 
   (define (warp-cursor-to-window w #!optional x y)
@@ -227,14 +254,18 @@ specified by the user."
 	  (setq x
 		(if (< (car warp-to-window-offset) 0)
 		    (car warp-to-window-offset)
-		  (quotient (* (car dims) (car warp-to-window-offset)) 100))))
+		  (quotient (* (car dims)
+                               (car warp-to-window-offset)) 100))))
 	(unless y
 	  (setq y (if (< (cdr warp-to-window-offset) 0)
 		      (cdr warp-to-window-offset)
-		    (quotient (* (cdr dims) (cdr warp-to-window-offset)) 100))))
+		    (quotient (* (cdr dims)
+                                 (cdr warp-to-window-offset)) 100))))
 	(warp-cursor
-	 (max 0 (min (1- (screen-width)) (+ x (car coords) (- (car foff)))))
-	 (max 0 (min (1- (screen-height)) (+ y (cdr coords) (- (cdr foff)))))))))
+	 (max 0 (min (1- (screen-width))
+                     (+ x (car coords) (- (car foff)))))
+	 (max 0 (min (1- (screen-height))
+                     (+ y (cdr coords) (- (cdr foff)))))))))
 
   (define (activate-window w)
     (require 'sawfish.wm.focus)
@@ -244,9 +275,9 @@ specified by the user."
     (when (window-really-wants-input-p w)
       (set-input-focus w))
     (warp-pointer-if-necessary w)
-    (window-order-push w))
+    (window-order-push w)
+    (call-window-hook 'activate-window-hook w))
 
-
 ;;; resizing windows in accordance with their size hints
 
   (define (constrain-dimension-to-hints x dimension hints)
@@ -264,12 +295,19 @@ specified by the user."
 		   (+ (* (ceiling (/ (- x bottom) inc)) inc) bottom)))))
       (clamp x (or minimum base 1) maximum)))
 
-  (define (resize-window-with-hints w cols rows #!optional hints)
-    "Resize window W to COLS x ROWS, using the window's size hints to define
-the row and column size, and the minimum possible size.
+  (define (constrain-aspect-to-hints a old-b dimension min-aspect max-aspect)
+    (let ((min-ratio (/ (cadr min-aspect) (cddr min-aspect)))
+          (max-ratio (/ (cadr max-aspect) (cddr min-aspect))))
+      (if (eq dimension 'y)
+          (clamp old-b (floor (* a min-ratio)) (floor (* a max-ratio)))
+        (clamp old-b (floor (/ a max-ratio)) (floor (/ a min-ratio))))))
 
-If HINTS is non-nil, then it is the size hints structure to use. Otherwise
-(window-size-hints W) is used."
+  (define (resize-window-with-hints w cols rows #!optional hints)
+    "Resize window W to COLS x ROWS, using the window's size hints to
+define the row and column size, and the minimum possible size.
+
+If HINTS is non-nil, then it is the size hints structure to
+use. Otherwise (window-size-hints W) is used."
     (unless hints
       (setq hints (window-size-hints w)))
     (let ((x-base (or (cdr (or (assq 'base-width hints)
@@ -283,13 +321,13 @@ If HINTS is non-nil, then it is the size hints structure to use. Otherwise
        (constrain-dimension-to-hints (+ y-base (* y-inc rows)) 'y hints))))
 
   (define (resize-window-with-hints* w width height #!optional hints)
-    "Resize window W to WIDTH x HEIGHT, with WIDTH and HEIGHT defined in
-terms of pixels. The window's size hints structure defines the minimum
-and maximum dimensions of the window, within which WIDTH and HEIGHT are
-constrained.
+    "Resize window W to WIDTH x HEIGHT, with WIDTH and HEIGHT defined
+in terms of pixels. The window's size hints structure defines the
+minimum and maximum dimensions of the window, within which WIDTH and
+HEIGHT are constrained.
 
-If HINTS is non-nil, then it is the size hints structure to use. Otherwise
-(window-size-hints W) is used."
+If HINTS is non-nil, then it is the size hints structure to
+use. Otherwise (window-size-hints W) is used."
     (unless hints
       (setq hints (window-size-hints w)))
     (resize-window-to w (constrain-dimension-to-hints width 'x hints)
@@ -345,7 +383,6 @@ If HINTS is non-nil, then it is the size hints structure to use. Otherwise
 	  (adjust-position-for-gravity/y
 	   w grav (cdr coords) #:inverse unadjust)))
 
-
 ;;; deleting windows
 
   (define (delete-window w #!optional safely)
@@ -365,7 +402,6 @@ If HINTS is non-nil, then it is the size hints structure to use. Otherwise
 
   (define-command 'delete-window-safely delete-window-safely #:spec "%W")
 
-
 ;;; making window names unique
 
   (define (uniquify-name in existing)
@@ -389,13 +425,13 @@ string `uniquify-name-format' to generate unique names."
 
   (define-command 'uniquify-window-name uniquify-window-name #:spec "%W")
 
-
-;; selecting a single window
+;;; selecting a single window
 
   (define select-window-map
     (bind-keys (make-keymap)
-      "Any-Click1" (lambda ()
-		     (throw 'select-window (query-pointer-window)))))
+               "Any-Click1" (lambda ()
+                              (throw 'select-window
+                                     (query-pointer-window)))))
 
   (defvar select-window-cursor-shape 'crosshair)
 
@@ -408,7 +444,6 @@ string `uniquify-name-format' to generate unique names."
 	      (recursive-edit)))
 	(ungrab-pointer))))
 
-
 ;;; avoided (i.e. non-overlapped) windows
 
   (define (window-avoided-p w)
@@ -428,7 +463,6 @@ returned in the list."
 		      (and (not (eq w window))
 			   (window-avoided-p w)))))
 
-
 ;;; property and window-state changed interface
 
   (define prop-changes '())
@@ -436,11 +470,25 @@ returned in the list."
   (define (call-after-property-changed prop fun)
     "Arrange for function FUN to be called with arguments (WINDOW PROPERTY
 STATE) when the X11 property named PROP (a symbol) changes. PROP may also
-be a list of property names to monitor."
-    (setq prop-changes (cons (cons (if (listp prop)
-				       prop
-				     (list prop)) fun) prop-changes)))
-  
+be a list of property names to monitor.
+
+Kluge: if PROP is `WM_NAME', it is replaced with `(WM_NAME _NET_WM_NAME)'.
+This is done to cope with themes that want to update title bars on name
+changes but only watch `WM_NAME'.  A warning is printed on stderr when
+this substitution occurs.  Those themes should really be fixed."
+    (setq prop-changes
+          (cons
+           (cons (cond ((listp prop) prop)
+                       ((eq prop 'WM_NAME)
+                        (format standard-error
+" (call-after-property-changed 'WM_NAME ...) should probably be
+ (call-after-property-changed '(WM_NAME _NET_WM_NAME) ...);
+Use '(WM_NAME) if you really want only WM_NAME\n")
+                        '(WM_NAME _NET_WM_NAME))
+                       (t (list prop)))
+                 fun)
+           prop-changes)))
+
   (add-hook 'property-notify-hook
 	    (lambda (w prop state)
 	      (mapc (lambda (cell)
@@ -449,7 +497,7 @@ be a list of property names to monitor."
 		    prop-changes)))
 
   (define state-changes '())
-  
+
   (define (call-after-state-changed states fun)
     "Arrange for function FUN to be called with arguments (WINDOW
 CHANGED-STATES) when one of the states defined by the list of symbols
@@ -457,20 +505,32 @@ STATES has been changed. STATES may also be a single symbol."
     (setq state-changes (cons (cons (if (listp states)
 					states
 				      (list states)) fun) state-changes)))
-  
+
   (add-hook
    'window-state-change-hook
    (lambda (w states)
      (mapc (lambda (cell)
-	     (let
-		  ((relevant (filter (lambda (state)
-					(memq state (car cell))) states)))
+	     (let ((relevant (filter (lambda (state)
+                                       (memq state (car cell))) states)))
 	       (when relevant
 		 ((cdr cell) w relevant))))
-	    state-changes)))
+           state-changes)))
 
-
-;; gaollable functions
+  (define (rename-window-func window new-name)
+    "Renames the WINDOW to NEW-NAME."
+    (set-x-text-property window 'WM_NAME (vector new-name))
+    (set-x-text-property window '_NET_WM_NAME (vector new-name))
+    (set-x-text-property window 'WM_ICON_NAME (vector new-name))
+    (set-x-text-property window '_NET_WM_ICON_NAME (vector new-name)))
+
+  (define (rename-window-interactive w)
+    (require 'sawfish.wm.util.prompt)
+    (let ((new-name (prompt-for-string "Enter new window title:"
+				       (window-name w))))
+      (rename-window-func w new-name)))
+
+  (define-command 'rename-window rename-window-interactive #:spec "%W")
+  ;;; gaollable functions
 
   (gaol-add window-really-wants-input-p window-class window-avoided-p
 	    call-after-property-changed call-after-state-changed
